#!/usr/bin/env python

from collections import defaultdict

from distributions import *
from schedcat.model.LLCLayout import *
from schedcat.model.LLC_params import *
import schedcat.model.ohc as ohc
import random
import time

from pulp import *
import gurobipy

import numpy as np

capped_wss = {  'B' : [512, 512, 512, 512],
                'C' : 512}
                
def LPLLCWays(ts_mcx, vartype, csch = bypass_scheme(), olap = True, debug = False, noIso = False):
#     print "vars:", len(ts_mcx), vartype, csch, olap, debug, noIso
    ts_mcx.setScheme(csch)
    ts_mcx.A = ts_mcx.level('A')
    ts_mcx.B = ts_mcx.level('B')
    ts_mcx.C = ts_mcx.level('C')
     
    #----------------------- Gather Data -----------------------------#
    UA_Adata = np.zeros((CORES, csch.mW+1))
    UA_Bdata = np.zeros((CORES, csch.mW+1))
    UA_Cdata = np.zeros((CORES, csch.mW+1))
    
    UB_Bdata = np.zeros((CORES, csch.mW+1))
    UB_Cdata = np.zeros((CORES, csch.mW+1))
    UC_Cdata = np.zeros(csch.mW+1)
    
    hdata = np.zeros(csch.mW+1)
    Hdata = np.zeros(csch.mW+1)
    
    minAperiods = []
    for core in range(0, 4):
        if len(ts_mcx.A.core(core)) == 0:
            minAperiods.append(0)
        else:
            minAperiods.append(min([t.period for t in ts_mcx.A.core(core)]))
    
    for w in range(0,csch.mW+1):
        wsch = scheme([area(w, csch.A[i].S) for i in range(0, CORES)],
                      [area(w, csch.B[i].S) for i in range(0, CORES)],
                      area(w, csch.C.S), csch.mW, csch.mS)
                      
        if noIso:
            wsch = scheme([area(0, csch.A[i].S) for i in range(0, CORES)],
                          [area(0, csch.B[i].S) for i in range(0, CORES)],
                          area(0, csch.C.S), csch.mW, csch.mS)
                          
        ts_mcx.setScheme(wsch)
        for core in range(0, CORES):
            UA_Adata[core][w] = ts_mcx.A.core(core).utilization('A', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UA_Bdata[core][w] = ts_mcx.A.core(core).utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UA_Cdata[core][w] = ts_mcx.A.core(core).utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            
            UB_Bdata[core][w] = ts_mcx.B.core(core).utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UB_Cdata[core][w] = ts_mcx.B.core(core).utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        
        UAi = defaultdict()
        UAi['B'] = [UA_Bdata[core][w] for core in range(0, CORES)]
        UAi['C'] = [UA_Cdata[core][w] for core in range(0, CORES)]
        
        UAi = ts_mcx.secondaryInflateConst(UAi, minAperiods = minAperiods)
        for core in range(0, CORES):
            UA_Bdata[core][w] = UAi['B'][core]
            UA_Cdata[core][w] = UAi['C'][core]
        
        UC_Cdata[w] = ts_mcx.C.utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        
        hdata[w] = ts_mcx.C.max_utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        Hdata[w] = ts_mcx.C.utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss, heaviest = CORES-1)
        
    
    #----------------------- Create LP Variables -----------------------------#
    problem = LpProblem("Hybrid",LpMinimize)
    
    UA_A = LpVariable.matrix("UA_A", range(CORES), lowBound = 0)
    UA_B = LpVariable.matrix("UA_B", range(CORES), lowBound = 0)
    UA_C = LpVariable.matrix("UA_C", range(CORES), lowBound = 0)
    
    UB_B = LpVariable.matrix("UB_B", range(CORES), lowBound = 0)
    UB_C = LpVariable.matrix("UB_C", range(CORES), lowBound = 0)
    
    UC_C = LpVariable("UC_C", lowBound = 0)
    
    h = LpVariable("h", lowBound = 0)
    H = LpVariable("H", lowBound = 0)
    
    Wa = LpVariable.matrix("Wa", range(CORES), lowBound = 0, cat=vartype)
    Wb = LpVariable.matrix("Wb", range(CORES), lowBound = 0, cat=vartype)
    Wc = LpVariable("Wc", lowBound = 0, cat=vartype);
    
    WO = LpVariable.matrix("WO", range(CORES), lowBound = 0)
    
    Wc.upBound = csch.mW
    
    
    #----------------------- Constraint Set 1 -----------------------------#
    for i, c in enumerate(Wa):
        Wa[i].upBound = csch.mW
        Wb[i].upBound = csch.mW
        WO[i].upBound = csch.mW
        
        if olap:
            problem += Wa[i] + Wc <= csch.mW
            problem += Wb[i] + Wc <= csch.mW
        else:
            problem += Wa[i] + Wb[i] + Wc <= csch.mW
            
        problem += WO[i] >= Wa[i] + Wb[i] + Wc - csch.mW
            
    #----------------------- Constraint Set 2 -----------------------------#
    for i in range(0, CORES):
        UA_A[i].upBound = 1
        UA_B[i].upBound = 1
        UA_C[i].upBound = 1
        
        UB_B[i].upBound = 1
        UB_C[i].upBound = 1
        
        if not len(ts_mcx.A.core(i)) is 0:
            invminp = 1.0/min([t.period for t in ts_mcx.A.core(i)])
        else:
            invminp = 0
            
        tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
        tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
        IA_B = ohc.CPMD*Wb[i]*csch.B[i].S*invminp*tckinfw if olap else 0
        IA_C = ohc.CPMDavg*WO[i]*csch.B[i].S*invminp*tckinfa if olap else 0
        
        for j in range(0,csch.mW):
            C1 = (UA_Adata[i][j+1]-UA_Adata[i][j])
            C2 = UA_Adata[i][j]-j*C1
            problem += (UA_A[i]-C1*Wa[i] >= C2)
            C1 = (UA_Bdata[i][j+1]-UA_Bdata[i][j])
            C2 = UA_Bdata[i][j]-j*C1
            problem += (UA_B[i]-C1*Wa[i] >= C2 + IA_B)
            C1 = (UA_Cdata[i][j+1]-UA_Cdata[i][j])
            C2 = UA_Cdata[i][j]-j*C1
            problem += (UA_C[i]-C1*Wa[i] >= C2 + IA_C)
            
            C1 = (UB_Bdata[i][j+1]-UB_Bdata[i][j])
            C2 = UB_Bdata[i][j]-j*C1
            problem += (UB_B[i]-C1*Wb[i] >= C2)
            C1 = (UB_Cdata[i][j+1]-UB_Cdata[i][j])
            C2 = UB_Cdata[i][j]-j*C1
            problem += (UB_C[i]-C1*Wb[i] >= C2)
    
    UC_C.upBound = CORES
    for w in range(0, csch.mW):
        C1 = (UC_Cdata[w+1]-UC_Cdata[w])
        C2 = UC_Cdata[w]-w*C1
        problem += (UC_C-C1*Wc >= C2)
        
    #----------------------- Constraint Set 3 -----------------------------#
        problem += h >= hdata[w] + (Wc-w)*(hdata[w+1]-hdata[w])
        problem += H >= Hdata[w] + (Wc-w)*(Hdata[w+1]-Hdata[w])
        
    #----------------------- Constraint Set 4 -----------------------------#
    
    for c in range(0,CORES):
        problem += UA_A[c] <= 1
        problem += UA_B[c] + UB_B[c] <= 1
        problem += UA_B[c] + UB_B[c]
        
    problem += lpSum(UA_C) + lpSum(UB_C) + UC_C <= CORES
        
    eps = 1e-6
    
    problem += lpSum(UA_C) + lpSum(UB_C) + (CORES-1)*h + H <= CORES-eps
    
    problem += lpSum(UA_C) + lpSum(UB_C) + UC_C
    
    #----------------------- Solve -----------------------------#
    status = problem.solve(GUROBI(msg=0, timeLimit = 120))

    if LpStatus[status] == 'Optimal':
        ps = scheme([area(Wa[i].varValue, csch.A[i].S) for i in range(0,CORES)],
                    [area(Wb[i].varValue, csch.B[i].S) for i in range(0,CORES)],
                    area(Wc.varValue, MAX_SETS), MAX_WAYS, MAX_SETS)
        
        ps.A = [val.rounded() for val in ps.A]
        ps.B = [val.rounded() for val in ps.B]
        ps.C = ps.C.rounded()
        
        return ps
        
def LPLLCWaysInter(ts_mcx, vartype, csch = bypass_scheme(), olap = True, debug = False, noIso = False):
#     print "vars:", len(ts_mcx), vartype, csch, olap, debug, noIso

    ts_mcx.setScheme(csch)
    ts_mcx.A = ts_mcx.level('A')
    ts_mcx.B = ts_mcx.level('B')
    ts_mcx.C = ts_mcx.level('C')
     
    #----------------------- Gather Data -----------------------------#
    UA_Adata = np.zeros((CORES, csch.mW+1))
    UA_Bdata = np.zeros((CORES, csch.mW+1))
    UA_Cdata = np.zeros((CORES, csch.mW+1))
    
    UB_Bdata = np.zeros((CORES, csch.mW+1))
    UB_Cdata = np.zeros((CORES, csch.mW+1))
    
    minAperiods = []
    for core in range(0, 4):
        if len(ts_mcx.A.core(core)) == 0:
            minAperiods.append(0)
        else:
            minAperiods.append(min([t.period for t in ts_mcx.A.core(core)]))
    
    wsch = scheme([area(0, csch.A[i].S) for i in range(0, CORES)],
                      [area(0, csch.B[i].S) for i in range(0, CORES)],
                      area(csch.mW, csch.C.S), csch.mW, csch.mS)
                          
    ts_mcx.setScheme(wsch)
    
    UC_Cdata = ts_mcx.C.utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss, model_cpd = True)
        
    hdata = ts_mcx.C.max_utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss, model_cpd = True)
    Hdata = ts_mcx.C.utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss, heaviest = CORES-1, model_cpd = True)
    
    for w in range(0,csch.mW+1):
        wsch = scheme([area(w, csch.A[i].S) for i in range(0, CORES)],
                      [area(w, csch.B[i].S) for i in range(0, CORES)],
                      area(csch.mW, csch.C.S), csch.mW, csch.mS)
                          
        ts_mcx.setScheme(wsch)
        for core in range(0, CORES):
#             print "UA_A:"
            UA_Adata[core][w] = ts_mcx.A.core(core).utilization('A', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
#             print "UA_B:"
            UA_Bdata[core][w] = ts_mcx.A.core(core).utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
#             print "UA_C:"
            UA_Cdata[core][w] = ts_mcx.A.core(core).utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            
#             print "UB_B:"
            UB_Bdata[core][w] = ts_mcx.B.core(core).utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
#             print "UB_C:"
            UB_Cdata[core][w] = ts_mcx.B.core(core).utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        
        UAi = defaultdict()
        UAi['B'] = [UA_Bdata[core][w] for core in range(0, CORES)]
        UAi['C'] = [UA_Cdata[core][w] for core in range(0, CORES)]
        
        UAi = ts_mcx.secondaryInflateConst(UAi, minAperiods = minAperiods)
        for core in range(0, CORES):
            UA_Bdata[core][w] = UAi['B'][core]
            UA_Cdata[core][w] = UAi['C'][core]
    
    #----------------------- Create LP Variables -----------------------------#
    problem = LpProblem("Hybrid",LpMinimize)
    
    UA_A = LpVariable.matrix("UA_A", range(CORES), lowBound = 0)
    UA_B = LpVariable.matrix("UA_B", range(CORES), lowBound = 0)
    UA_C = LpVariable.matrix("UA_C", range(CORES), lowBound = 0)
    
    UB_B = LpVariable.matrix("UB_B", range(CORES), lowBound = 0)
    UB_C = LpVariable.matrix("UB_C", range(CORES), lowBound = 0)
    
    Wa = LpVariable.matrix("Wa", range(CORES), lowBound = 0, cat=vartype)
    Wb = LpVariable.matrix("Wb", range(CORES), lowBound = 0, cat=vartype)
    Wc = LpVariable("Wc", lowBound = 0, cat=vartype);
    
    WO = LpVariable.matrix("WO", range(CORES), lowBound = 0)
    
    Wc.upBound = csch.mW
    
    
    #----------------------- Constraint Set 1 -----------------------------#
    for i, c in enumerate(Wa):
        Wa[i].upBound = csch.mW
        Wb[i].upBound = csch.mW
        WO[i].upBound = csch.mW
        
        if olap:
            problem += Wa[i] + Wc <= csch.mW
            problem += Wb[i] + Wc <= csch.mW
        else:
            problem += Wa[i] + Wb[i] + Wc <= csch.mW
            
        problem += WO[i] >= Wa[i] + Wb[i] + Wc - csch.mW
            
    #----------------------- Constraint Set 2 -----------------------------#
    for i in range(0, CORES):
        UA_A[i].upBound = 1
        UA_B[i].upBound = 1
        UA_C[i].upBound = 1
        
        UB_B[i].upBound = 1
        UB_C[i].upBound = 1
        
        if not len(ts_mcx.A.core(i)) is 0:
            invminp = 1.0/min([t.period for t in ts_mcx.A.core(i)])
        else:
            invminp = 0
            
        tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
        tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
        IA_B = ohc.CPMD*Wb[i]*csch.B[i].S*invminp*tckinfw if olap else 0
        IA_C = ohc.CPMDavg*WO[i]*csch.B[i].S*invminp*tckinfa if olap else 0
        
        for j in range(0,csch.mW):
            C1 = (UA_Adata[i][j+1]-UA_Adata[i][j])
            C2 = UA_Adata[i][j]-j*C1
            problem += (UA_A[i]-C1*Wa[i] >= C2)
            C1 = (UA_Bdata[i][j+1]-UA_Bdata[i][j])
            C2 = UA_Bdata[i][j]-j*C1
            problem += (UA_B[i]-C1*Wa[i] >= C2 + IA_B)
            C1 = (UA_Cdata[i][j+1]-UA_Cdata[i][j])
            C2 = UA_Cdata[i][j]-j*C1
            problem += (UA_C[i]-C1*Wa[i] >= C2 + IA_C)
            
            C1 = (UB_Bdata[i][j+1]-UB_Bdata[i][j])
            C2 = UB_Bdata[i][j]-j*C1
            problem += (UB_B[i]-C1*Wb[i] >= C2)
            C1 = (UB_Cdata[i][j+1]-UB_Cdata[i][j])
            C2 = UB_Cdata[i][j]-j*C1
            problem += (UB_C[i]-C1*Wb[i] >= C2)
        
    #----------------------- Constraint Set 4 -----------------------------#
    
    for c in range(0,CORES):
        problem += UA_A[c] <= 1
        problem += UA_B[c] + UB_B[c] <= 1
        problem += UA_B[c] + UB_B[c]
        
    problem += lpSum(UA_C) + lpSum(UB_C) + UC_Cdata <= CORES
        
    eps = 1e-6
    
    problem += lpSum(UA_C) + lpSum(UB_C) + (CORES-1)*hdata + Hdata <= CORES-eps
    
    problem += lpSum(UA_C) + lpSum(UB_C) + UC_Cdata
    
    #----------------------- Solve -----------------------------#
    status = problem.solve(GUROBI(msg=0, timeLimit = 120))

    if LpStatus[status] == 'Optimal':
        ps = scheme([area(Wa[i].varValue, csch.A[i].S) for i in range(0,CORES)],
                    [area(Wb[i].varValue, csch.B[i].S) for i in range(0,CORES)],
                    area(16, MAX_SETS), MAX_WAYS, MAX_SETS)
        
        ps.A = [val.rounded() for val in ps.A]
        ps.B = [val.rounded() for val in ps.B]
        ps.C = ps.C.rounded()
        
        return ps